---- UI Interface.
-- Only available when native code was loaded.
--
-- Always check for presence before use.
--
-- This describes UI Version: 20200303
-- @author Oskar Eisemuth
-- @copyright 2018-2019
-- @module commonapi.ui
local ui = {
	_windowlist = {},
	_menulist = commonapi.utils.createOrderedTable(),
	_idgen = 1,
}

require "serialize"

local _utils = commonapi.utils
local _log = _utils.logPrep("commonapi._nativeui")

local uiscale = 1


--_log("Loading...")

local imgui = commonapi._native.getGUIBindings()
local UIEvent = require("commonapi2.uicommon.UIEvent")



--- UI interface
-- @section ui

--- UI Element types
-- @section uielementtypes

---
-- Common to all ui elements
--
--
--
-- @table UIElement
-- @tfield string typ Type of element (Label, Button, Separator, Checkbox ...)
-- @tfield[opt] bool visible
-- @field[opt] data User defined data (To carry state as example)
-- @see Separator
-- @see Button
-- @usage
-- {
--  typ = "Button",
--  id = "myButton1",
--  caption = "Button1",
-- }


---
-- @table Separator

--- Simple text label
-- @table Label
-- @tfield string caption

--- Checkbox Control
-- @table Checkbox
-- @tfield string id unique id
-- @tfield string caption
-- @tfield bool default
-- @tfield[opt] function onclick onclick(item, container)

--- Text Input Control
-- @table TextInput
-- @tfield string id unique id
-- @tfield string caption
-- @tfield[opt] string default



--- Button Control
--
--
--
-- Warning: container.params[id] will be set to true (this may change in the future)
-- @table Button
-- @tfield string id unique id
-- @tfield string caption
-- @tfield[opt] function onclick onclick(item, container)
-- @usage
-- {
--   typ = "Button",
--   id = "myButton1",
--   caption = "Button1",
--   onclick = function(item, container)
--     container:close()
--   end,
-- }


--- Float Input Control
-- @table FloatValue
-- @tfield string id unique id
-- @tfield[opt] float default
-- @tfield[opt] float step
-- @tfield[opt] float decimalPlaces
-- @tfield[opt] float minValue
-- @tfield[opt] float maxValue


--- Integer Input Control
-- @table IntValue
-- @tfield string id unique id
-- @tfield[opt] int default
-- @tfield[opt] int step
-- @tfield[opt] int minValue
-- @tfield[opt] int maxValue



---
-- @table ComboBox
-- @tfield string id unique id
-- @tfield table values List of ComboBox Values
-- @tfield[opt] int defaultIndex
-- @tfield[opt] function onclick onclick(item, container)




local UIItemRender = {

}

local function UIItemRender_Flags(flagconstants, flags, setflagstable)
	if (type(setflagstable) == 'table') then
		for k, v in pairs(setflagstable) do
			if (flagconstants[k] ~= nil) then
				if (v == true) then
					flags = bit32.bor(flags, flagconstants[k])
				else
					flags = bit32.band(flags, bit32.bnot(flagconstants[k]))
				end
			end
		end
	else
		flags = setflagstable
	end
	return flags
end

function UIItemRender.Separator(self, item)
	imgui.Separator()
	return false
end

function UIItemRender.Label(self, item)
	if (item.caption == nil) then
		return false
	end
	if (item.color ~= nil) then
		local color = item.color
		imgui.TextColored(color[1], color[2], color[3], color[4], "".. item.caption)
	else
		imgui.TextUnformatted("".. item.caption)
	end
	return false
end

function UIItemRender.Text(self, item)
	if (item.text == nil) then
		return false
	end
	imgui.PushTextWrapPos(0.0)
	if (item.color ~= nil) then
		local color = item.color
		imgui.TextColored(color[1], color[2], color[3], color[4], "".. item.text)
	else
		imgui.TextUnformatted("".. item.text)
	end
	imgui.PopTextWrapPos();
	return false
end


function UIItemRender.Checkbox(self, item)
	local changed
	local param

	if (self.params[item.id] ~= nil) then
		param = self.params[item.id]
	else
		param = item.default or false
	end
	changed, param = imgui.Checkbox("".. (item.caption or "") .. '##' .. item.id, param)
	if changed then
		self.params[item.id] = param

		UIEvent.createClick(self, item)
		--item._evt_clicked = true
	end
	return changed
end

function UIItemRender.TextInput(self, item)
	local changed = false
	local param = ""
	local value
	local ok
	
	if (self.params[item.id] ~= nil) then
		param = self.params[item.id]
	else
		param = item.default or ""
	end

	if (item.width ~= nil) then
		imgui.PushItemWidth(item.width*uiscale);
	else
		imgui.PushItemWidth(-1);
	end

	local flags = 0

	if (self.changeonenter == true) then
		flags = imgui.constant.InputTextFlags.EnterReturnsTrue
	end
	changed, param = imgui.InputText('##' .. item.id, param, 255, flags);
	imgui.PopItemWidth();
	if changed then
		self.params[item.id] = param
		UIEvent.createChange(self, item)
	end
	return changed
end


function UIItemRender.Button(self, item)
	local changed = false
	local nameid = (item.caption or '').. '##' .. item.id
	if (imgui.Button(nameid)) then
		self.params[item.id] = true
		--item._evt_clicked = true
		UIEvent.createClick(self, item)
		changed = true
	end
	return changed
end


function UIItemRender.SmallButton(self, item)
	local changed = false
	local nameid = (item.caption or '').. '##' .. item.id
	if (imgui.SmallButton(nameid)) then
		self.params[item.id] = true
		--item._evt_clicked = true
		UIEvent.createClick(self, item)
		changed = true
	end
	return changed
end

function UIItemRender.Buttons(self, item)
	local changed = false
	local param
	
	local i = 0
	local first = true

	if (self.params[item.id] ~= nil) then
		param = self.params[item.id]
	else
		param = item.defaultIndex or 0
	end

	if type(item.values) == "table" then
		for k, v in pairs(item.values) do
			if (first) then
				first = false
			else
				imgui.SameLine()
			end

			if (param == i) then
				imgui.PushStyleColor(imgui.constant.Col.Button, 0xFFFF0000)
				imgui.PushStyleColor(imgui.constant.Col.ButtonActive, 0xFFFF0000)
				imgui.PushStyleColor(imgui.constant.Col.ButtonHovered, 0xFFFF0000)
			end
			imgui.PushID(item.id)
			local textid = v..'##'..item.id..i
			if (imgui.Button(textid)) then
				self.params[item.id] = i
				--item._evt_clicked = i
				UIEvent.createClick(self, item)
				changed = true
			end
			imgui.PopID()
			if (param == i) then
				imgui.PopStyleColor(3)
			end

			i = i + 1
		end
	end

	return changed
end

function UIItemRender.FloatValue(self, item)
	local changed = false
	local param = 0
	local step = 0.1
	local decimalPlaces = 2
	local formatPlaces = "%.2f"

	if (self.params[item.id] ~= nil) then
		param = self.params[item.id]
	else
		param = item.default or 0
	end

	if (item.step ~= nil) then
		step = item.step
	end
	if (item.decimalPlaces ~= nil) then
		decimalPlaces = item.decimalPlaces
		formatPlaces = "%."..decimalPlaces.."f"
	end

	--imgui.PushItemWidth(-5);
	imgui.PushItemWidth(-10)
	changed, param = imgui.InputFloat("##"..item.id, param, step, step, formatPlaces);
	imgui.PopItemWidth();
	if (changed) then
		local mult = 10^(decimalPlaces or 0)
		local param = math.floor(param * mult + 0.5) / mult
		if (item.minValue ~= nil) then
			if (param < item.minValue) then
				param = item.minValue;
			end
		end
		if (item.maxValue ~= nil) then
			if (param > item.maxValue) then
				param = item.maxValue;
			end
		end

		self.params[item.id] = param
	end
	return changed
end


function UIItemRender.IntValue(self, item)
	local changed = false
	local param = 0
	local step = 0.1

	if (self.params[item.id] ~= nil) then
		param = self.params[item.id]
	else
		param = item.default or 0
	end

	if (item.step ~= nil) then
		step = item.step
	end

	imgui.PushItemWidth(-10)
	changed, param = imgui.InputInt("##"..item.id, param, step, step);
	imgui.PopItemWidth();
	if (changed) then
		if (item.minValue ~= nil) then
			if (param < item.minValue) then
				param = item.minValue;
			end
		end
		if (item.maxValue ~= nil) then
			if (param > item.maxValue) then
				param = item.maxValue;
			end
		end
		self.params[item.id] = param
	end
	return changed
end

function UIItemRender.ComboBox(self, item)
	local changed = false
	local param = 0
	local clicked = false
	
	local i = 0
	local first = true

	if (self.params[item.id] ~= nil) then
		param = self.params[item.id]
	else
		param = item.defaultIndex or 0
	end
	
	if type(item.values) == "table" then

		if (item.values[param+1] == nil) then
			param = 0
		end
		imgui.PushItemWidth(-10)
		if (imgui.BeginCombo("##"..item.id, item.values[param+1], flags)) then
			i = 0
			for key, value in ipairs(item.values) do
				local is_selected = (i == param);
				clicked, is_selected = imgui.Selectable(value, is_selected)
				if clicked then
					self.params[item.id] = i
					changed = true
					--item._evt_clicked = i
					UIEvent.createClick(self, item)
				end
				i = i + 1;
				if (is_selected) then
					imgui.SetItemDefaultFocus()
				end
			end
			imgui.EndCombo();
		end
		imgui.PopItemWidth();
	end

	return changed
end

function UIItemRender._Selectable(self, item)
	local changed = false
	local nameid = (item.caption or '').. '##' .. item.id
	local isSelected

	local flags = imgui.constant.SelectableFlags.AllowDoubleClick
	if (item.flags ~= nil) then
		flags = UIItemRender_Flags( imgui.constant.SelectableFlags, flags, item.flags)
	end

	local selectedType = type(item.selected)
	if (selectedType == 'function') then
		isSelected = item.selected(item, self)
	else
		isSelected = (item.selected or false)
	end
	if (item.color ~= nil) then
		local color = item.color
		imgui.PushStyleColor_2(imgui.constant.Col.Text, color[1], color[2], color[3], color[4])
	end
	if (imgui.Selectable(nameid, isSelected, flags)) then
		self.params[item.id] = true
		if (imgui.IsMouseDoubleClicked(0)) then
			--item._evt_dblclicked = true
			UIEvent.createDblClick(self, item)
		else
			--item._evt_clicked = true
			UIEvent.createClick(self, item)
		end
		changed = true
	end
	if (item._fix_overlap == true) then
		imgui.SetItemAllowOverlap()
	end
	if (item.color ~= nil) then
		imgui.PopStyleColor(1)
	end
	return changed
end


function UIItemRender.Columns(self, item)
	if (item.count ~= nil) then
		imgui.Columns(item.count, nil, (item.border == true) )
	else
		imgui.Columns(1)
	end

	return false
end



function UIItemRender.Table(self, item)
	local changed = false
	local flags = imgui.constant.TableFlags.Resizable + imgui.constant.TableFlags.NoSavedSettings
	--flags = imgui.constant.TableFlags.Resizable + imgui.constant.TableFlags.BordersOuter + imgui.constant.TableFlags.BordersV
	if (item.flags ~= nil) then
		flags = UIItemRender_Flags(imgui.constant.TableFlags, flags, item.flags)
	end
	local isOpen
	if (item.size ~= nil) then
		flags = flags + imgui.constant.TableFlags.ScrollY
		isOpen = imgui.BeginTable('##'..item.id, item.columns or 1, flags, item.size[1] * uiscale, item.size[2] * uiscale);
	else
		isOpen = imgui.BeginTable('##'..item.id, item.columns or 1, flags);
	end
	if (isOpen) then
		changed = self:_renderItems(item.uilist, true)
		imgui.EndTable()
	end
	return changed
end

function UIItemRender.TableHeader(self, item)
	local changed = false
	local items = item.items or {}
	for k, header in ipairs(items) do
		UIItemRender.TableSetupColumn(self, header)
	end

	imgui.TableAutoHeaders()
	item.cursorting = imgui.TableGetSortSpecs()
	if (item.cursorting ~= nil and item.cursorting.changed) then
		changed = true
		UIEvent.createChange(self, item)
	end
	return changed
end




function UIItemRender.TableSetupColumn(self, item)
	local flags = 0 --imgui.constant.TableColumnFlags.Resizable
	if (item.flags ~= nil) then
		flags = UIItemRender_Flags(imgui.constant.TableColumnFlags, flags, item.flags)
	end
	if (item.caption == nil) then
		return false
	end
	if (item.width ~= nil) then
		flags = bit32.bor(flags, imgui.constant.TableColumnFlags.WidthFixed)
		imgui.TableSetupColumn(item.caption, flags, item.width*uiscale);
	else
		imgui.TableSetupColumn(item.caption, flags);
	end
	return false
end
function UIItemRender.TableHeadersDone(self, item)
	local changed = false
	imgui.TableAutoHeaders()
	local sortspecs = imgui.TableGetSortSpecs()
	if (sortspecs ~= nil and sortspecs.changed) then
		changed = true
		UIEvent.createChange(self, item)
	end
	return changed
end



function UIItemRender.ScrollRegion(self, item)
	local changed
	local flags
	flags = imgui.constant.WindowFlags._HorizontalScrollbar
	imgui.BeginChild(item.id, item.size[1] * uiscale, item.size[2] * uiscale, item.border == true, flags);
	changed = self:_renderItems(item.uilist)
	imgui.EndChild()
	return changed
end

function UIItemRender._ScrollRegionStart(self, item)
	local flags
	flags = imgui.constant.WindowFlags._HorizontalScrollbar
	imgui.BeginChild(item.id, item.size[1] * uiscale, item.size[2] * uiscale, item.border == true, flags);
	return false
end


function UIItemRender._ScrollRegionEnd(self, item)
	imgui.EndChild()
	return false
end



function UIItemRender.TabBar(self, item)
	local changed = false
	local flags
	local nameid = (item.caption or '').. '##' .. item.id
	flags = 0 -- imgui.constant.WindowFlags._HorizontalScrollbar
	if (imgui.BeginTabBar(item.id, flags)) then
		changed = self:_renderItems(item.uilist or {})
		imgui.EndTabBar()
	end
	
	return changed
end


function UIItemRender.TabItem(self, item)
	local changed = false
	local flags
	local nameid = (item.caption or '').. '##' .. item.id
	flags = 0 -- imgui.constant.WindowFlags._HorizontalScrollbar
	if (imgui.BeginTabItem(nameid)) then
		changed = self:_renderItems(item.uilist or {})
		imgui.EndTabItem()
	end
	return changed
end


function UIItemRender._Popup(self, item)
	local changed = false

	if (self.params[item.id] == true) then
		imgui.OpenPopup(item.id)
		self.params[item.id] = nil
	end

	if (imgui.BeginPopup(item.id)) then
		changed = self:_renderItems(item.uilist)
		imgui.EndPopup()
	end
	return changed
end

--- UIWindow Class
--
-- Provides a user programmable window
--
-- @usage
-- local mywin = commonapi.ui.Window:new()
-- mywin.id = "mymodid_mwywin"
-- mywin.caption = "Samplewindow"
-- mywin:open()
--
-- @type UIWindow
local UIWindow = {
	_open = true,

	id = "",
	uilist = {},
	params = {},
	["_"] = commonapi.utils._t
}

--- UI Window creation function
-- @tparam[opt] table o A table to use while construction
-- @treturn UIWindow
function UIWindow:new(o)
	local o = o or {}

	for k, v in pairs(self) do
		if (o[k] == nil) then
			o[k] = _utils.deepcopy(v)
		end
	end

	--o._super = self
	return o
end

function UIWindow:newClass(o)
	local o = o or {}

	for k, v in pairs(self) do
		if (o[k] == nil) then
			o[k] = _utils.deepcopy(v)
		end
	end

	o._super = self
	return o
end



--- Unique id
--- @tfield string id Id of Window

--- Elements of window
--- @tfield {UIElement,...} uilist List of ui elements

--- Paramater
--- @tfield table params Parameters by ui element.id, may only contain modified parameters

--- onchange callback
---
--- will be fired when params change (by ui element)
--- @tfield function|nil onchange function onchange(win)

--- Width of window
--- @tfield[opt] int width Width of window

--- Height of window
-- @tfield[opt] int height Height of window



--- Get element by id
-- @tfield string id
-- @return UIElement
-- @return[1] nil
function UIWindow:getElementById(id)
	for k, item in pairs(self.uilist) do
		if (item.id == id) then
			return item
		end
	end
	return nil
end



function UIWindow:_renderItems(uilist, tableapi)
	local changed = false
	for k, item in pairs(uilist) do
		local typ = item.typ or "Label"

		if (tableapi == true) then
			if (item.rownew == true) then
				imgui.TableNextRow()
			end

			if (item.column ~= nil) then
				imgui.TableSetColumnIndex(item.column)
			end
		else
			if (item.column ~= nil) then
				local curcolumn = imgui.GetColumnIndex()
				if (curcolumn ~= item.column) then
					if ( item.column < imgui.GetColumnsCount() ) then
						local maxcount = 0
						while ( curcolumn ~= item.column) do
							imgui.NextColumn()
							curcolumn = imgui.GetColumnIndex()
							maxcount = maxcount + 1
							if (maxcount > 10) then
								--print("imgui.NextColumn() logic error!")
								--print(debug.traceback())
								break
								--error("imgui.NextColumn() logic error!")
							end
						end
					end
				elseif (item.rownew == true) then
					imgui.NextColumn()
					curcolumn = imgui.GetColumnIndex()
					local maxcount = 0
					while ( curcolumn ~= item.column) do
						imgui.NextColumn()
						curcolumn = imgui.GetColumnIndex()
						maxcount = maxcount + 1
						if (maxcount > 100) then
							--print("imgui.NextColumn() logic error!")
							--print(debug.traceback())
							break
							--error("imgui.NextColumn() logic error!")
						end
					end
				end
				if (self._firstframe) then
					if (item.columnwidth ~= nil) then
						imgui.SetColumnWidth(curcolumn, item.columnwidth * uiscale)
					end
					if (item.columnoffset ~= nil) then
						imgui.SetColumnOffset(curcolumn, item.columnoffset * uiscale)
					end
				end
			end
		end
		if (item.visible ~= false) then
			if (item.sameline ~= nil) then
				imgui.SameLine()
			end
			local fn = UIItemRender[typ]
			if (fn(self, item)) then
				changed = true
			end
		end
	end
	return changed
end

--
function UIWindow:_render()
	local shoulddraw 
	local changed = false
	local imguidid
	local imguiflags = 0
	if (self.autoresize ~= false) then
		imguiflags = imgui.constant.WindowFlags.AlwaysAutoResize
	end

	if (type(self.caption) == "string") then
		imguidid = self.caption.. "##" .. self.id
	else
		imguidid = "Window ".. self.id .. "##" .. self.id
	end

	if (self.width ~= nil or self.height ~= nil) then
		imguiflags = 0 --imgui.constant.WindowFlags.AlwaysAutoResize
		--imguiflags = imgui.constant.WindowFlags.NoResize

		local w = (self.width or 0) * uiscale
		local h = (self.height or 0) * uiscale
		imgui.SetNextWindowSize(w, h, imgui.constant.Cond.Appearing)
	end
	shoulddraw, self._open = imgui.Begin( imguidid , self._open, imguiflags)
	if (self._open == false) then
		imgui.End()
		self:close()
		return
	end
	if (shoulddraw == false) then
		imgui.End()
		return
	end

	changed = self:_renderItems(self.uilist)

	if (changed) then
		self.changed = true
	end
	if (self._firstframe) then
		self._firstframe = false
	end
	imgui.End()
end

function UIWindow:_fireevents()
	local win = self
--[[
	for k, item in pairs(self.uilist) do
		if (item._evt_clicked ~= nil) then
			if (item.onclick ~= nil) then
				item.onclick(item, win)
--			elseif (win.onclick ~= nil) then
--				win.onclick(self, item)
			end
			item._evt_clicked = nil
		elseif (item._evt_dblclicked ~= nil) then
			if (item.ondblclick ~= nil) then
				item.ondblclick(item, win)
			end
			item._evt_dblclicked = nil
		end
	end
--]]

	UIEvent.fireEvents(self)

	if (self.changed == true) then
		if (self.onchange ~= nil) then
			self:onchange();
		end
		self.changed = false
	end
end


--- Open this window
--
--
--
-- Warning: Proper id of window needs to be set before use!
function UIWindow:open()
	self._open = true
	self._firstframe = true
	ui._addWindow(self)
end

--- Close this window
function UIWindow:close()
	self._open = false
	UIEvent.createClose(self, self)
	ui._removeWindow(self)
end


----
--- UI interface
-- @section ui


--- UI Window
-- @tparam UIWindow Window UI Window Class
ui.Window = UIWindow

---
-- Get Window by id
-- @tparam string id
-- @treturn UIWindow|nil
function ui.getWindow(id)
	return ui._windowlist[id]
end

function ui._removeWindow(obj)
	local id = rawget(obj, 'id')
	if (id ~= nil) then
		ui._windowlist[id] = nil
	end
end

function ui._addWindow(obj)
	local id = rawget(obj, 'id')
	if (id == nil) then
		ui._idgen = ui._idgen + 1
		obj.id = ui._idgen;
	end
	ui._windowlist[obj.id] = obj
end

function ui._render(typ)
	uiscale = commonapi._native.getUIScale();
	if (typ == 1) then
		ui._fireevents()
		ui._renderwindowlist()
		return
	end

	if (typ == 2) then
		ui._renderMenu()
		return
	end
end

function ui._renderMenu()
	if (imgui.BeginMenu("Mods##ModsMenu")) then
		for id, menuobj in pairs(ui._menulist) do
			if (imgui.MenuItem(menuobj.caption .. '##'..id)) then
				menuobj._evt_clicked = true
			end
		end
		imgui.EndMenu();
	end
end





function ui._renderwindowlist()
	local ok, err, remove

	for key, windowobj in pairs(ui._windowlist) do
		ok, err = pcall(windowobj._render, windowobj)
		if (ok == false) then
			_log("Rendering failed for id: ", key)
			_log(err)
			remove = key
		end
	end
	if (remove ~= nil) then
		ui._windowlist[remove] = nil
	end
end

function ui._fireevents()
	local ok, err

	for key, menuobj in pairs(ui._menulist) do
		if (menuobj._evt_clicked == true) then
			menuobj._evt_clicked = false
			--menuobj.fn(key)
			ok, err = pcall(menuobj.fn, key)
			if (ok == false) then
				_log("Fireing events failed for menu id: "..key.."\n"..err);
			end
		end
	end

	local idlist = {}
	for id, windowobj in pairs(ui._windowlist) do
		idlist[#idlist+1] = id
	end

	for idx, id in pairs(idlist) do
		local windowobj = ui._windowlist[id]
		if (windowobj ~= nil) then
			ok, err = pcall(windowobj._fireevents, windowobj)
			if (ok == false) then
				_log("Fireing events failed for id: ", id, err);
				ui._windowlist[id] = nil
			end
		end
	end
--[[
	local remove = nil
	for id, windowobj in pairs(ui._windowlist) do
		ok, err = pcall(windowobj._fireevents, windowobj)
		if (ok == false) then
			_log("Fireing events failed for id: ", id, err);
			remove = id
		end
		--windowobj:_fireevents();
	end

	if (remove ~= nil) then
		ui._windowlist[remove] = nil
	end
--]]
end


function ui._activate()
	commonapi._native.setGUIRenderFn(ui._render)
end

---
-- Creates a entry in the commonapi mod menu
-- @tparam string id unique id for this entry
-- @tparam string caption
-- @tparam function fn callback function
function ui.createModMenuEntry(id, caption, fn)
	ui._menulist[id] = {
		caption = caption,
		fn = fn,
		_evt_clicked = false
	}
end

---
-- Removes id from commonapi mod menu
function ui.removeModMenuEntry(id)
	--ui._menulist[id] = nil
	ui._menulist:_delete(id)
end

--- Message Box Creator
-- @tparam string id unique id for this msgbox. (overwrites previous message with same id)
-- @tparam string text
-- @tparam string caption
function ui.msgbox(id, text, caption, buttons)
	local uilist = {}
	local msgbox = ui.Window:new()

	msgbox.id = "msgbox_"..id
	msgbox.caption = caption


	uilist[#uilist+1] = {
		typ = "Label",
		caption = text
	}

	if (buttons == nil) then
		uilist[#uilist+1] = {
			typ = "Button",
			caption = _("OK"),
			id = "ok",
			onclick = function(item, win)
				win:close()
			end
		}
	else
		for i, entry in ipairs( buttons ) do
			uilist[#uilist+1] = {
				typ = "Button",
				id = entry['caption'],
				caption = entry['caption'],
				data = {
					fn = entry['fn'] or false
				},
				onclick = function(item, win)
					win:close()
					if (item.data.fn) then
						item.data.fn()
					end
				end
			}
		end
	end

	msgbox.uilist = uilist
	msgbox:open()
end





local function reloadnativecallback()
	print("Reload nativeui callback called")
	package.loaded["commonapi._nativeui"] = nil
	commonapi._nativeui = nil
	commonapi.ui = commonapi._nativeui
	commonapi.ui._activate()
end
--ui.createModMenuEntry('nativeui_reload', "Native UI Reload", reloadnativecallback)

return ui